'''
Runner functions are responsible for running some specific interaction example
and gathering input, ouput and prompt strings.
'''

from judge import errors
from judge import builtins
from judge import util
from judge.util import timeout as run_with_timeout
from judge.parsers import Example, IOTemplate


def run_unsafe_python(example: Example, src: str, *,
                      namespace=None,
                      exec=exec,  # @ReservedAssignmenth
                      timeout=None):
    '''Run a string of Python code and return the Example() object for this
    interaction.

    Parameters
    ----------

    example : Example or template string
        The example object describing the interaction.
    src : str
        A string of Python source code or a code object that is passed to
        the exec() function.
    namespace : dict
        A namespace dictionary that is passed to the exec function.
    exec : callable
        A function that is executed as exec(src, namespace) in order to run
        the code in the given namespace. The function do not need to return
        anything, but rather must execute the correct `print` and `input`
        statements in order to recriate the correct Example() output.
    '''
    if isinstance(example, str):
        example = Example(example)

    # Prepare list of inputs and interactions
    interaction = []
    inputs = list(example.get_inputs())
    namespace = namespace if namespace is not None else {}

    # Prepare overrides for the builtin functions
    print_func = util.make_interaction_print(interaction)
    input_func = util.make_interaction_input(interaction, inputs)

    # Run code and return interactions. All errors that are not related to
    # code giving wrong answers will be captured at this stage
    builtins.update(print=print_func, input=input_func)
    try:
        if timeout:
            run_with_timeout(exec, args=(src, namespace), timeout=timeout)
        else:
            exec(src, namespace)

    # Execution exceeded timeout
    except TimeoutError:
        ex = errors.TimeExceededError()
        out = Example(interaction, error=ex, validate=False)

    # Program asked for a non-existing input
    except errors.MissingInputError as ex:
        out = Example(interaction, error=ex, validate=False)

    # Bad syntax in code
    except SyntaxError as ex:
        tb = ex.__traceback__
        ex = errors.BuildError(src, ex.with_traceback(tb))
        out = Example(interaction, error=ex, validate=False)

    # Some generic runtime error happened
    except Exception as ex:
        tb = ex.__traceback__
        ex = errors.UnexpectedError(ex.with_traceback(tb))
        out = Example(interaction, error=ex, validate=False)

    # Execution terminated
    else:
        # ... but did not used all necessary inputs
        if inputs:
            ex = errors.UnusedInputError(inputs)
            out = Example(interaction, error=ex, validate=False)
        else:
            out = Example(interaction, value=example.value,
                          feedback=example.feedback)

    finally:
        builtins.restore()
    return out


def run_python(example, src, **kwds):
    '''Safe sandboxed version of run_unsafe_python().

    This function has the overhead of starting another Python interpreter with
    reduced permissions.'''

    return run_unsafe_python(example, src, **kwds)


def run_unsafe_file(example: Example, path):
    '''Run executable file in the given path through shell and obtain the
    correct interactions'''

    raise NotImplementedError


def run_file(example: Example, path, **kwds):
    '''Safe sandboxed version of run_unsafe_file().

    This function runs the file with reduced permissions.'''

    return run_unsafe_file(example, path, **kwds)


def run_template_python(template, src, *,
                        stop_at_error=True,
                        check_value=True,
                        is_safe=False, **kwds):
    '''Run all examples in template and return the corresponding IOTemplate'''

    if isinstance(template, str):
        template = IOTemplate(template)

    # Choose runner function
    runner_func = run_unsafe_python if is_safe else run_python

    def runner(ex):
        result = runner_func(ex, src, **kwds)
        if isinstance(check_value, dict):
            return result.with_grades(ex, **check_value)
        elif check_value:
            return result.with_grades(ex)
        else:
            return result

    # Run examples and grade
    if stop_at_error:
        max_value = 1
        out = []
        for example in template:
            result = runner(example)
            out.append(result)
            if result.error:
                break
        return IOTemplate(out, value=max_value)
    else:
        return IOTemplate([runner(example) for example in template])


if __name__ == '__main__':
    import doctest
    doctest.testmod()

    code1 = "x=input('x'); y=input('y'); print(x + y)"
    code1 = "x=input('x'); y=input('y'); print('ab')"
    code2 = "x=input('x'); print('abc')"
    code3 = "x=input('x'); y=input('y'); z=input('z'); print(x + y)"
    tmpl1 = "x: a; y: b --> ab"
    tmpl2 = "x: 1; y: 2 --> 12"
    tmpl3 = "x: x; y: y --> xy"
    template = '\n\n'.join([tmpl1, tmpl2, tmpl3])

    print(run_python(tmpl1, code1))

    from judge.parsers import parse_io_template
    print(parse_io_template(template))

    out = run_template_python(template, code1)
    print(out)



from .io_example import ListLike

__all__ = ['IOTemplate']


class IOTemplate(ListLike):
    '''Represent a group of examples.

    It has a similar API to a list, but has some additional methods.
    '''

    def __repr__(self):
        tname = type(self).__name__
        if self.template is not None:
            return '%s(%r)' % (tname, self.template)

        optargs = ''
        if self.value != 1:
            optargs += ', value=' + str(self.value)
        # if self.feedback:
        #    optargs += ', feedback=' + repr(self.feedback)

        return '%s(%r%s)' % (tname, list(self), optargs)

    def _parse(self, template):
        from .io_parser import parse_io_template
        return parse_io_template(template)

    @property
    def grade(self):
        return self.value * min(ex.grade for ex in self)

    def with_grades(self, template):
        '''Return a graded copy of itself'''

        if isinstance(template, str):
            template = IOTemplate(template)

        data = [x.with_grades(y) for (x, y) in zip(self, template)]
        return IOTemplate(data, value=self.value)

    def get_grade(self):
        '''Returns the grade for the given run'''

        return self.grade

    def get_message(self, format='text'):  # @ReservedAssignment
        '''Return a formatted message for the given run'''

        if self.grade != 1:
            mingrade = min(self, key=lambda x: x.grade)
            return mingrade.get_message()



import io
import sys
import signal
import functools
import html
from threading import Thread

_print_func = print

#
# String formmating
#


def indent(st, indent=4):
    '''Indent string by the given number of spaces or the given indentation
    string'''

    indent = ' ' * indent if isinstance(indent, int) else indent
    lines = [indent + x for x in st.splitlines()]
    return '\n'.join(lines)


def html_escape(x):
    '''Escape unsafe HTML characters such as < > & etc'''

    return html.escape(x)


#
# Special IO functions for python script interactions
#
def capture_print(*args, **kwds):
    '''A print-like function that return the formatted string instead of
    printing it on screen'''

    out, err = sys.stdout, sys.stderr
    out_io = sys.stdout = sys.stderr = io.StringIO()
    try:
        _print_func(*args, **kwds)
    finally:
        sys.stdout, sys.stderr = out, err
    return out_io.getvalue()


def make_interaction_print(interaction):
    '''Returns a print() function that instead of printing through the regular
    channels, it appends an Out string to the given list of interactions'''

    from .parsers import Out

    @functools.wraps(print)
    def print_func(*args, **kwds):
        data = capture_print(*args, **kwds)
        if interaction and isinstance(interaction[-1], Out):
            data = interaction.pop() + data
        interaction.append(Out(data))

    return print_func


def make_interaction_input(interaction, inputs):
    '''Returns an input() function that instead of asking for user input
    through the regular channels, it retrives values from the inputs list and
    saves the corresponding interaction in the interactoin list.'''

    from . import errors
    from .parsers import In, Prompt

    @functools.wraps(input)
    def input_func(msg=None):
        if msg:
            interaction.append(Prompt(msg))
        try:
            data = inputs.pop(0)
        except IndexError:

            raise errors.MissingInputError
        interaction.append(In(data))
        return data

    return input_func

#
# Execution with timeout limits
#


def __timeout_handler(signum, frame):
    '''Helper function for the timeout() function.'''

    raise TimeoutError()


def timeout(func, args=(), kwargs={}, timeout=1.0, threading=True):
    '''Execute callable `func` with timeout.

    If timeout exceeds, raises a TimeoutError'''

    if threading:
        result = []
        exceptions = []

        def target():
            try:
                result.append(func(*args, **kwargs))
            except Exception as ex:
                exceptions.append(ex)

        thread = Thread(target=target)
        thread.start()
        thread.join(timeout)
        if thread.is_alive():
            raise TimeoutError
        else:
            try:
                return result.pop()
            except IndexError:
                raise exceptions.pop()
    else:
        signal.signal(signal.SIGALRM, __timeout_handler)
        signal.alarm(timeout)
        try:
            result = func(*args, **kwargs)
        finally:
            signal.alarm(0)

        return result



'''
Created on 29/09/2015

@author: chips
'''
import os
import argparse
from collections import namedtuple
from judge import runners

Grade = namedtuple('Grade', ['value', 'message'])


def parser():
    '''Returns the parser object for script.'''

    parser = argparse.ArgumentParser(
        description='Automatically grade the input file')
    parser.add_argument('file', help='input file to grade')
    parser.add_argument('--template', '-t',
                        help='grader template file')
    parser.add_argument('--grade', '-g', action='store_true',
                        help=('simply return the final grade without '
                              'showing messages'))
    return parser


def grade_file(src_path, template_path):
    '''Return a Grade object corresponding to the given grading job'''

    with open(template_path) as F:
        template = F.read()

    with open(src_path) as F:
        src = F.read()
        src_ext = os.path.splitext(src_path)[1]

    if src_ext == '.py':
        result = runners.run_template_python(template, src)
        result = result.with_grades(template)
        grade = result.get_grade()
        message = result.get_message(format='text')
        return Grade(grade, message)
    else:
        raise ValueError(
            0,
            'file with extension %s is not supported' %
            src_ext)


def main():
    '''Executes the main script'''

    # Fetch parameters from argparse
    args = parser().parse_args()
    src = args.file
    template = args.template
    grade_only = args.grade
    if template is None:
        template = os.path.splitext(src)[0] + '.io'

    try:
        grade = grade_file(src, template)
    except Exception as ex:
        # Handle errors
        if ex.args[0] == 0:
            print('Error:', ex)
            raise SystemExit
        else:
            value = input('An unknown error was found, '
                          'do you wish to see the traceback? [y/N] ')

            ex = ex if value.lower() in ['y', 'yes'] else SystemExit
            raise ex

    if not grade_only and grade.message:
        print(grade.message)
    if grade.value == 1:
        print('Final grade: 100%, congratulations!')
    else:
        print('Final grade: %s%%' % (grade.value * 100))



import copy
import collections
from decimal import Decimal
from .. import errors
from .io_strings import *


__all__ = ['Example']


class ListLike(collections.Sequence):
    '''Base class for Example and IOTemplate objects'''

    def __init__(self, template, value=1, parsed=None):
        # Fix template vs. parsed list input
        if isinstance(template, list):
            parsed, template = list(template), None
        elif parsed is None:
            parsed = self._parse(template)

        self.template = template
        self._data = list(parsed)
        self.value = Decimal(value)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = Decimal(value)

    def _parse(self, template):
        raise NotImplementedError

    def __getitem__(self, idx):
        return self._data[idx]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __eq__(self, other):
        return (len(self) == len(other) and
                all(x == y for (x, y) in zip(self, other)))


class Example(ListLike):
    '''An example of input/output/prompt interactions'''

    _loose_match_args = dict(strip=True)

    def __init__(self, template, value=None, feedback='', parsed=None,
                 error=None, validate=True):
        if value is None:
            value = 0 if error else 1
        self.feedback = feedback
        self.error = error
        super().__init__(template, value=value, parsed=parsed)

        # Some consistency checks
        if validate:
            assert self.__has_one_input_per_prompt() or self.is_input

    def _parse(self, template):
        from .io_parser import parse_example
        return parse_example(template)

    #
    # Properties
    #
    @property
    def is_input(self):
        return all(isinstance(x, In) for x in self)

    @property
    def is_output(self):
        return all(isinstance(x, Out) for x in self)

    @property
    def is_prompt(self):
        return all(isinstance(x, Prompt) for x in self)

    @property
    def grade(self):
        return self.value

    #
    # Public API
    #
    def get_inputs(self):
        '''Return a list with all input strings.'''

        return [x for x in self if isinstance(x, In)]

    def get_outputs(self, *, prompts=True, convert=False):
        '''Return a list with all output strings, including prompts.

        If `prompts=False`, skip Prompt strings from output. If
        `convert=True`, convert Prompt strings to Out strings
        '''

        if prompts:
            data = [x for x in self if isinstance(x, (Out, Prompt))]
            if convert:
                data = [Out(x) for x in data]
            return data
        else:
            return [x for x in self if isinstance(x, Out)]

    def get_grade(self):
        '''Returns the grade for the given run'''

        return self.grade

    def get_message(self, format='text'):  # @ReservedAssignment
        '''Return a formatted message for the given run'''

        if self.error:
            return self.error.message(format)

    def loose_match(self, other, strip=False, casefold=False):
        r'''Match two example objects allowing small (but well defined)
        discrepancies. The default behavior is similar to ``self == other``
        but additional parameters may allow specific types of discrepancies.


        Parameters
        ----------

        strip : bool
            If True, ignore all trailing newlines and whitespace for the last
            output value
        casefold : bool
            If True, convert all strings to lowercase before comparison.


        Examples
        --------

        >>> ex1 = Example('x:a; y:b --> ab')
        >>> ex2 = Example('x:a; y:b --> "ab "')

        These two examples are slightly different since the second outputs an
        additional trailing space.

        >>> ex1 == ex2
        False

        >>> ex1.loose_match(ex2, strip=True)
        True
        '''

        A = list(self)
        B = list(other)

        # Apply transformations
        if strip:
            if isinstance(A[-1], Out):
                A[-1] = A[-1].strip()
            if isinstance(B[-1], Out):
                B[-1] = B[-1].strip()
        if casefold:
            A = [x.casefold() for x in A]
            B = [x.casefold() for x in B]

        # Test othe matches
        is_match = A == B

        return is_match

    def with_grades(self, template):
        '''Return a copy of itself with any possible grading errors.'''

        if isinstance(template, str):
            template = Example(template)

        if self.error or self.loose_match(template, **self._loose_match_args):
            new = self.copy()
        else:
            new = self.copy()
            new.value = 0
            new.error = errors.WrongAnswerError(self, template)

        new.template = template
        return new

    def copy(self):
        '''Return a shallow copy of itself'''

        new = copy.copy(self)
        new._data = new._data[:]
        return new

    def format(self, method='text'):
        try:
            method = getattr(self, 'format_' + method)
        except AttributeError:
            raise ValueError('invalid method: %r' % method)
        return method()

    def format_text(self):
        msgs = []
        for x in self:
            if isinstance(x, In):
                msgs.append('< %s >\n' % x)
                # msgs.append('\n')
            elif isinstance(x, Prompt):
                msgs.append(str(x))
            elif isinstance(x, Out):
                msgs.append(str(x) + '\n')
        return (''.join(msgs)).rstrip()

    #
    # Magic methods
    #
    def __repr__(self):
        if self.template is not None:
            return 'Example(%r)' % self.template

        data = list(self)
        optargs = ''
        if self.value != 1:
            optargs += ', value=' + str(self.value)
        if self.feedback:
            optargs += ', feedback=' + repr(self.feedback)
        if self.error:
            optargs += ', error=' + repr(self.error)

        return 'Example(%r%s)' % (data, optargs)

    #
    # Consistency checks
    #
    def __has_one_input_per_prompt(self):
        n_prompts = sum(1 for x in self if isinstance(x, Prompt))
        n_inputs = sum(1 for x in self if isinstance(x, In))
        return n_prompts == n_inputs



'''IO questions templates specify a sequence of input/output pairs expected
to appear in a given program. It is a simple language that is parsed into a
sequence of Example() sequences that are suitable to be used as a list of
examples by an automated grades. Optionally, the user may specify the input
strings and a program that compute the responses as output strings.

Template language
-----------------

The most simple construct is the input/output pair given bellow::

    "input prompt": "input string" --> "output string"

The quotes are optional, but may be necessary to escape special character
sequeces such as ":", "-->", etc, or to have a precise control of whitespace.
Consecutive lines are interpreted as a sequence of interations whithin a single
execution. Use a blank line to separate different executions. Comments start
with "#".

::
    # First example: sum two floats
    "x: ": "1.0";
    "y: ": "2.0";
    --> "sum = 3.0"

    # Second example: sum two ints
    "x: ": "1";
    "y: ": "2";
    --> "sum = 3.0"

Since the colon character ``:`` often appears in input prompts, we created the
following shortcuts: ``x::y`` is translated to ``"x:": "y"`` and ``x:: y`` is
translated to ``"x: ": "y"``. This allows some compact representations such
as::

    # sum two floats
    x:: 1.0; y:: 2.0 --> sum = 3.0

    # sum two ints
    x:: 1; y:: 2 --> sum = 3.0

Multiline output strings can be specified by a heading ``|`` in the beginning
of each line

::
    x:: a; y:: b;
    -->
        |multiline output
        |another line
        |another line

Indentation is optional but helps with readability.


If only input strings are specified, the grader may compute the complete
interaction from an example program which represent a correct answer. In this
case, just list inputs as a sequence of strings followed by semi-colons::

    1; 2; 3; 4;

New lines and other forms of whitespace are not important.

Each block of examples can start with a parenthesized argument list that
specify some options to the grader. The grader accept two optional parameters
as in the example::

    # sum two ints
    x:: 1; y:: 2 --> sum = 3.0

    # An argument modifying a single example
    [ 50%, output must display a least a decimal place ]
    x:: 1; y:: 2 --> sum = 3

    # An argument modifying multiples examples
    [ 75%, output must have the same number of decimal places as the input ]
        x:: 1.25; y:: 1.25 --> sum = 2.50

        x:: 1.250; y:: 1.25 --> sum = 2.500

        x:: 1.25; y:: 1.250 --> sum = 2.500


Configuration options
---------------------


Parsed output
-------------

...
'''

import re
from . import In, Out, Prompt, Example, IOTemplate

__all__ = ['parse_io_template']


def parse_io_template(template):
    '''Parse the template string into a parsed list of examples.

    Example
    -------

    >>> parse_io_template('x:a; y:b --> ab')
    IOTemplate([Example('x:a; y:b --> ab')])

    Each example consists of a list of input/output/prompt strings

    >>> example = parse_io_template('x:a; y:b --> ab')[0]
    >>> list(example)
    [Prompt('x'), In('a'), Prompt('y'), In('b'), Out('ab')]
    '''

    template = '\n'.join([line.strip() for line in template.splitlines()])
    examples = [line.strip() for line in template.split('\n\n') if line]
    examples = [parse_example(example) for example in examples]
    return IOTemplate([x for x in examples if x is not None])


def parse_example(template):
    '''Parse a string for a single example'''

    tk_list = tokenizer(template)

    # Join string lists
    while True:
        try:
            idx = tk_match_seq(tk_list, String, Whitespace, String)
            st = tk_list[idx][Example('x:a; y:b --> ab')]
            st += tk_list.pop(idx + 1)
            st += tk_list.pop(idx + 1)
            tk_list[idx] = String(st)
        except ValueError:
            break

    # Strip comments and whitespace
    tk_list = [tk for tk in tk_list if not isinstance(tk, Comment)]
    tk_list = [tk for tk in tk_list if not isinstance(tk, Whitespace)]

    # Return an empty template
    if not tk_list:
        return None

    while isinstance(tk_list[0], (Paragraph, NewLine)):
        del tk_list[0]
    while isinstance(tk_list[-1], (Paragraph, NewLine)):
        tk_list.pop()

    # Convert multiline strings
    for idx, tk in enumerate(tk_list):
        if isinstance(tk, MultiLine):
            data = ''
            for line in tk[1:-1].splitlines():
                data += line.partition('|')[-1] + '\n'
            tk_list[idx] = String(data[:-1])

    # Match foo:: bar and convert to "foo: ": bar
    while True:
        try:
            idx = tk_match_seq(tk_list, String, (':: ', '::'), String)
            sep = tk_list[idx + 1]
            tk_list[idx] = String(tk_list[idx] + sep[1:])
            tk_list[idx + 1] = Colon(':')
        except ValueError:
            break

    # Match all arrows and insert output
    while True:
        try:
            idx = tk_match_seq(tk_list, Arrow, String, (NewLine, SemiColon))
            tk_list.pop(idx + 2)
        except ValueError:
            break

    while True:
        try:
            idx = tk_match_seq(tk_list, Arrow, String)
            data = tk_list.pop(idx + 1)
            tk_list[idx] = Out(data)
        except ValueError:
            break

    # Match all colon definitions
    while True:
        try:
            idx = tk_match_seq(tk_list, String, Colon, String)
            in_ = In(tk_list.pop(idx + 2))
            out = Prompt(tk_list[idx])
            tk_list[idx] = out
            tk_list[idx + 1] = in_
        except ValueError:
            break

    # Match all String/semicolon combinations
    while True:
        try:
            idx = tk_match_seq(tk_list, String, (NewLine, SemiColon))
            data = tk_list.pop(idx)
            tk_list[idx] = In(data)
        except ValueError:
            break

    # Match all input_str/semicolon combinations
    while True:
        try:
            idx = tk_match_seq(tk_list, In, (NewLine, SemiColon))
            tk_list.pop(idx + 1)
        except ValueError:
            break

    if isinstance(tk_list[0], Args):
        args, tk_list = tk_list[0], tk_list[1:]
        args, kwargs = args
        kwargs['parsed'] = tk_list
        return Example(template, *args, **kwargs)
    return Example(template, parsed=tk_list)


#
# Input stream tokenizer
#
class Token(str):

    '''Base class for all tokenizer in the input stream'''

    def __repr__(self):
        tname = type(self).__name__
        return '%s(%s)' % (tname, super().__repr__())


class Whitespace(Token):
    regex = re.compile(r'[ \t]+')


class String(Token):
    regex = re.compile(r'''("[^"\n]*"|'[^'\n]*')''')


class Colon(Token):
    regex = re.compile(r'(:: |::|:)')


class SemiColon(Token):
    regex = re.compile(r';')


class NewLine(Token):
    regex = re.compile(r'(\r\n|\n)')


class Arrow(Token):
    regex = re.compile(r'-->')


class MultiLine(Token):
    regex = re.compile(r'(\n[ ]+[|][^\n]*)+')


class Comment(Token):
    regex = re.compile(r'#[^\n]*\n')


class Paragraph(Token):
    regex = re.compile(r'([ \t]*(\r\n|\n))+')


TK_CLASSES = [Whitespace, String, Colon, SemiColon, NewLine, Arrow, MultiLine,
              Comment, Paragraph]
REGEX_PAIRS = [(cls.regex, cls) for cls in TK_CLASSES]
REGEX_MAP = dict(REGEX_PAIRS)
REGEX_LIST = [x for (x, y) in REGEX_PAIRS]

#
# Utility functions
#


def tokenizer(template, out=None):
    '''Return a string of tokens from the given input template string.'''

    out = [] if out is None else out
    if not template.endswith('\n'):
        template += '\n'

    while template:
        # Match any regex expression
        matches = [re.search(template) for re in REGEX_LIST]
        matches = [(m.start(), m.end(), m) for m in matches if m is not None]
        if any(matches):
            # Group by lowest starting index
            matches.sort(key=lambda x: x[0])
            lowest = matches[0][0]
            matches = [(j, m) for (i, j, m) in matches if i == lowest]

            # Find largest match in group
            matches.sort(key=lambda x: x[0], reverse=True)
            highest, match = matches[0]

            # Extract match, data, etc
            pre = template[:lowest]
            data = template[lowest:highest]
            template = template[highest:]

            # Process string tokenizer
            if match.re is String.regex:
                data = data[1:-1]

            # Check if pre has invalid characters
            # ...
            if pre:
                out.append(String(pre))
            out.append(REGEX_MAP[match.re](data))
            continue

        template = out.append(template)

    return out


def tk_match(value, arg):
    '''Matches value with argument either by type (e.g.:
    ``isinstance(value, arg)``), or by value (e.g.: ``value == arg``)'''

    if isinstance(arg, type):
        return isinstance(value, arg)
    else:
        return arg == value


def tk_match_seq(L, *args, raises=True):
    '''Return the index of the first token at a sequence that matches each of
    the given extra arguments::


    tk_match_seq(L, String, '::', String) --> the first ocurrence of tokens
    that matches each of the given arguments in order.
    '''

    n = len(L) - len(args) + 1
    for i, x in enumerate(L[:n]):
        for j, arg in enumerate(args):
            value = L[i + j]
            if not isinstance(arg, (tuple, list)):
                arg = (arg,)
            if not any(tk_match(value, x) for x in arg):
                break
        else:
            return i

    if raises:
        raise ValueError
    else:
        return -1


def tk_split(tk_list, sep):
    '''Split token list in the ``head, sep, tail`` parts.'''

    if isinstance(sep, type):
        def split(x):
            return isinstance(x, sep)
    else:
        def split(x):
            return x == sep

    group = []
    groups = [group]

    for tk in tk_list:
        if split(tk):
            group = []
            groups.append(group)
        else:
            group.append(tk)
    return groups


class Args(object):
    '''Represent the arguments passed to the example constructor.'''

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        yield self.args
        yield self.kwargs


if __name__ == '__main__':
    f1 = '''
    # foo
    x:y;x:y-->bar

    x: y; x:y --> bar
    x:y --> foo
    '''

    f2 = 'foo x: y --> bar'

    f3 = '''foo:: 3
    -->
        |foo
        |bar
        |    foobar'''

    f4 = '-->"hello world"'
    f = f1

    print(parse_io_template(f))

    import doctest
    doctest.testmod()



from decimal import Decimal
from codeschool.graders import errors
from codeschool.graders.util import timeout, message_from_exception


class GraderTemplate(object):

    '''
    :cls:`GraderTemplate` objects are use to grade code submited to some
    specific coding problem.
    '''

    def __init__(self, *, answer=None, max_runtime=None):
        self.answer = answer
        self.max_runtime = max_runtime

    def validate(self, code):
        '''Raise a `BadSyntaxError` if the string of code has some syntax
        error.
        '''

        raise NotImplementedError

    def grade(self, code):
        '''Run grading and return a :cls:`Grade` object with the grade
        associated to code.'''

        raise NotImplementedError

    def timed_grading(self, code, time=1):
        '''Run grading job with a timeout'''

        try:
            return timeout(self.grade, args=(code,), timeout=time)
        except TimeoutError:
            ex = errors.TimeExceededError()
            return Decimal(0), message_from_exception(ex)

    def __call__(self, code, timeout=None):
        if timeout is not None:
            return self.timed_grading(code, timeout)
        else:
            return self.grade(code)


class PythonTemplate(GraderTemplate):

    '''
    A validator that assumes a Python code input.

    Users can specify a pre-processor that changes the input code prior
    processing. The preprocessor can insert imports, block some language
    features or perform arbitrary code transformations.
    '''
    _preprocessors = {}

    def __init__(self, *, preprocessor=lambda x: x, **kwds):
        D = self._preprocessors
        self.preprocessor = D.get(preprocessor, preprocessor)
        super().__init__(**kwds)

    def validate(self, code):
        code = self.preprocessor(code)
        return compile(code, '<input>', 'exec')

    def execute(self, code):
        '''Execute some Python code and return the resulting namespace.'''

        code_obj = self.validate(code)
        namespace = {}
        exec(code_obj, namespace)
        return namespace

    @classmethod
    def register_preprocessor(cls, name, value=None):
        '''Register a pre-processor function.

        A registered preprocessor can be referenced by name in the
        instantiation of a PythonTemplate object.

        Example
        -------

        >>> from pytuga import transpile
        >>> PythonTemplate.register_preprocessor('pytuga', transpile)
        '''
        # Decorator form
        if value is None:
            def decorator(func):
                cls.register_preprocessor(name, func)
                return func
            return decorator

        cls._preprocessors[name] = value



import os
import contextlib
import pwd
import subprocess
from subprocess import CalledProcessError
from fs.tempfs import TempFS
from django.utils.html import escape as escape_html


@contextlib.contextmanager
def tempdir(path):
    '''Temporarely sets the working directory to the given working_dir.
    Restores the original working_dir after leaving the with block.
    '''

    old_path = os.getcwd()
    try:
        os.chdir(path)
        yield None
    finally:
        os.chdir(old_path)


class Message(object):

    '''Represents a message that can be rendered using different methods.'''

    def __init__(self, data, success=False):
        self.data = data
        self.success = success

    def render(self, method='string'):
        '''Render message as 'string' or 'html', depending on the chosen
        method'''

        if method == 'html':
            return self._render_html()
        elif method == 'string':
            return self._render_string()
        else:
            raise ValueError('invalid render method: %s' % method)

    def _reder_html(self):
        return escape_html(self.render_string())

    def _render_string(self):
        return str(self.data)

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self.data)


class ErrorMessage(Message):

    def __init__(self, data):
        super().__init__(data, success=False)

    def _render_string(self):
        ex = self.data
        if isinstance(self.data, CalledProcessError):
            data = ex.output
            data += 'Error: exited with code %s' % ex.returncode
        else:
            data = '%s: %s' % (type(ex).__name__, ex)
        return data


class ExceptionMessage(Message):
    pass
    # def _render_html(self):
    #    pass


class Sandbox(object):

    '''
    Creates a generic sandbox environment that can be used to run arbitrary
    executables.

    The sandboxed job consists of three phases: preparation, build and
    execution.

        1) The first phase is used to insert files or set up any configuration
           options. This phase should always succeed irrespective of user
           input. Tipically it involves copying files and preparing the build
           environment.
        2) The second phase uses the user provided information and build or
           prepares an executable. This phase can throw syntax errors, but
           should not be vunerable to malicious attack. Tipically it involves
           compiling or simply checking source files for syntax errors.
        3) The actual runtime: run the code supplied by the user with several
           different examples and check the inputs and outputs in each case.
           This phase run the user supplied programs in a sandboxed
           environment by providing a very restricted user account.

    Notice that no code should depend on absolute paths, since they will run in
    some arbitrary working_dir in the filesystem.

    The approach took here sets the correct UID and GID and only works in
    POSIX systems.
    '''
    _vars = ['working_dir', 'is_prepared', 'timeout', 'username', 'password',
             'uid', 'gid', 'source_filename', 'build_command', 'run_command']

    def __init__(self, timeout=1.0):
        self.working_dir = TempFS()
        self.is_prepared = False
        self.timeout = timeout
        self.username = 'user_a'
        self.password = 'pipoca77'  # very bad?
        self._info = pwd.getpwnam(self.username)
        self.uid = self._info.pw_uid
        self.gid = self._info.pw_gid

    def vars(self):
        '''Return a dictionary of attributes that are used to compose values
        in user strings'''

        return {k: getattr(self, k, None) for k in self._vars}

    def syspath(self, path=''):
        '''Return the system working_dir for the temporary directory.

        This is useful for interacting with system functions. Creation and
        manipulation of files are better done through the TempFS object stored
        in the ``working_dir`` attribute.
        '''

        return self.working_dir.getsyspath(path)

    def prepare(self):
        '''Prepare all files for execution in an isolated environment.

        Return a Message object in case of failure or None in case of
        success.

        Subclasses must override the "_prepare" method that is called by this
        method.'''

        try:
            self._prepare()
        except Exception as ex:
            fmt = '%s(%s)' % (type(ex).__name__, ex)
            raise RuntimeError(
                ('Exception caught when preparing environment: %s\n'
                 'The _prepare() method must never raise exceptions.') % fmt)
        else:
            self._prepare_permissions()
            self.is_prepared = True

    def _prepare_permissions(self):
        '''Run chmod to change the ownership of all files in the prepared
        working directory'''

        os.chmod(self.syspath(), 0x755)

        return
        fs = self.working_dir
        paths = list(fs.walk())
        paths.reverse()

        for path, files in paths:
            for file in files:
                fpath = fs.getsyspath(file)
                os.chown(fpath, self.uid, self.gid)
            os.chown(path, self.uid, self.gid)

    def build(self):
        '''Build the executable (or simply check for syntax errors).

        Return a Message object in case of failure or None in case of
        success.

        Subclasses must override the "_build" method that is called by this
        method.'''

        with tempdir(self.working_dir.getsyspath('')):
            return self._build()

    def execute(self, inputs=()):
        '''Execute code passing the given sequence of input strings.

        Return a Message object that can represent a success of failure.

        Subclasses must override the "_execute" method that is called by this
        method.'''

        with tempdir(self.working_dir.getsyspath('')):
            return self._execute(inputs)

    _prepare = _build = lambda self: None
    _execute = lambda _self, inputs: None

    def _call_unprivileged(self, cmd, inputs, env={}, **kwds):
        '''Spawn a subprocess using the given command that runs in one of the
        unpriviledged "user_a", "user_b", etc accounts'''

        _env, env = env, dict(
            HOME=self.working_dir.getsyspath(''),
            LOGNAME=self.username,
            USER=self.username,
            PWD=self.working_dir.getsyspath(''),
        )
        env.update(_env)

        def preexec_fn():
            '''Configure process to run in the correct UID and GID'''

            info = pwd.getpwnam('nobody')
            os.setegid(info.pw_gid)
            os.seteuid(info.pw_uid)

        return subprocess.check_output(
            cmd,
            input=inputs,
            universal_newlines=True,
            stderr=subprocess.STDOUT,
            timeout=self.timeout,
            # preexec_fn=preexec_fn,
            shell=True,
            env=env,
            **kwds)


class SourceSandbox(Sandbox):

    '''A sandboxed job based on the user sending a single file of source code
    for evaluation.

    Sub-classes must implement specific languages.
    '''

    source_filename = 'source.txt'
    build_command = None
    run_command = './{source_filename}'

    def __init__(self, source):
        self.source = source
        super().__init__()

    def _prepare(self):
        with self.working_dir.open(self.source_filename, 'w') as F:
            # F.write((
            #    'import os\n'
            #    'os.setuid(%s)\n'
            #    'os.setgid(%s)\n') % (self.uid, self.gid))
            F.write(self.source)
        #path = self.syspath(self.source_filename)

    def _build(self):
        os.system(self.build_command.format(**self.vars()))

    def _execute(self, inputs):
        cmd = self.run_command.format(**self.vars())
        inputs = self.password + '\n' + '\n'.join(inputs)
        try:
            output = self._call_unprivileged(cmd, inputs)
        except subprocess.CalledProcessError as ex:
            ex.output = ex.output[10:]
            return ErrorMessage(ex)
        else:
            return Message(output[10:], success=True)


class PyScriptSandbox(SourceSandbox):
    source_filename = 'source.py'
    build_command = None
    run_command = 'su {username} -c "python source.py"'

    def _build(self):
        try:
            compile(self.source, self.source_filename, 'exec')
        except SyntaxError as ex:
            return Message(ex)


class PytugaScriptSandbox(PyScriptSandbox):
    source_filename = 'source.pytg'
    build_command = None
    run_command = 'su {username} -c "pytuga source.pytg"'

    def _build(self):
        import pytuga
        src = pytuga.transpile(self.source)
        try:
            compile(src, self.source_filename, 'exec')
        except SyntaxError as ex:
            return Message(ex)


if __name__ == '__main__':
    src = "print('hello world')"
    #src = 'print("sum:", float(input("x: ")) + float(input("y: ")))'
    #src = 'x = 1\nwhile x:\n  x = input()\n  print(x)'
    #src = 'raise ValueError'
    job = PytugaScriptSandbox(src)
    job = PyScriptSandbox(src)
    job.prepare()
    job.build()
    print(job.execute(['1', '2']).render())



from django.utils.html import escape
from codeschool.graders import errors
_print_func = print


def example_to_html(example):
    '''Render a parsed IO template into HTML'''

    out = []
    if isinstance(example, str):
        return escape(example)
    for x in example:
        data = x.replace('\n', '<br>')  # TODO: proper HTML escaping
        if x.is_input:
            data = '<span style="color: #f00;">&lt;%s&gt;</span>' % data
            data += '<br>'
        out.append(data)
    return ''.join(out)  # + '<pre>%s</pre>' % (example)


def message_bad_format(got, expected):
    got = '<strong>Valor obtido</strong><br>' + example_to_html(got)
    expected = '<strong>Resposta esperada</strong><br>' + \
        example_to_html(expected)
    return '<div>%s<br><br>%s</div>' % (got, expected)


def message_from_exception(ex):
    return str(type(ex).__name__ + ': ' + str(ex))



'''
In, Out, and Prompt strings for Example objects.
'''

import collections

__all__ = ['In', 'Out', 'Prompt']


class ParsedString(collections.UserString):

    '''Strings that appear in the parsed output'''

    is_input = False
    is_output = False
    is_prompt = False

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, str(self))

    def render(self):
        '''Return a string representation of the token'''

        return str(self)


class In(ParsedString):

    '''Represent input strings in an example interaction.'''

    is_input = True


class Out(ParsedString):

    '''Represent output strings in an example interaction.'''

    is_output = True


class Prompt(Out):

    '''Represent a string that is shown when asking the user for some input.
    The grader may interpret these strings as optional.'''

    is_prompt = True



#-*- coding: utf8 -*-
import os
import sys
import setuptools
from setuptools import setup

#
# Read VERSION from file and write it in the appropriate places
#
AUTHOR = 'Fábio Macêdo Mendes'
BASE, _ = os.path.split(__file__)
with open(os.path.join(BASE, 'VERSION')) as F:
    VERSION = F.read().strip()
with open(os.path.join(BASE, 'src', 'judge', 'meta.py'), 'w') as F:
    F.write(
        '# Auto-generated file. Please do not edit\n'
        '__version__ = %r\n' % VERSION +
        '__author__ = %r\n' % AUTHOR)

VERSION_BIG = VERSION.rpartition('.')[0]

#
# Main configuration script
#
setup(
    name='pyjudge',
    version=VERSION,
    description='An automatic judge for grading code',
    author='Fábio Macêdo Mendes',
    author_email='fabiomacedomendes@gmail.com',
    url='https://github.com/fabiommendes/pyjudge',
    long_description=open(os.path.join(BASE, 'README.txt')).read(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries',
    ],

    package_dir={'': 'src'},
    packages=setuptools.find_packages('src'),
    license='GPL',
    zip_safe=True,
    install_requires=[],
    entry_points={
        'console_scripts': ['pyjudge = judge.bin.judge:main'],
    },
)



import functools
from decimal import Decimal
from codeschool.parsers import parse_io_template, In, Out, Prompt
from codeschool.graders import errors
from codeschool.graders.base import PythonTemplate
from codeschool.graders.util import print_str, message_bad_format, message_from_exception


class PyIOGraderTemplate(PythonTemplate):

    '''
    A template that specifies the expected sequence of input/output entries.

    The code is run and the final grade is computed from the ratio of correct
    outputs.
    '''

    def __init__(self, examples, *,
                 strict_caps=True, fuzzy_matching=True, **kwds):
        super().__init__(**kwds)

        if isinstance(examples, str):
            examples = parse_io_template(examples)
        self.examples = list(examples)
        self.fuzzy_matching = fuzzy_matching
        self.strict_caps = strict_caps

        if self.answer.strip():
            self.answer_code = self.validate(self.answer)
            self.examples = [self.check_answer(ex) for ex in self.examples]

    def check_answer(self, example):
        if all(isinstance(atom, In) for atom in example):
            inputs = [str(x) for x in example]
            return self.interact(self.answer_code, inputs)
        else:
            grade, msgs = self.grade_example(self.answer_code, example)
            if grade != 100:
                raise ValueError('bad_example', msgs)
        return example

    def namespace(self):
        '''Return a new namespace for execution'''

        import tugalib
        return vars(tugalib)

    def grade(self, code):
        try:
            code_obj = self.validate(code)
        except SyntaxError:
            ex = errors.BadSyntaxError()
            return Decimal(0), message_from_exception(ex)

        out = Decimal(100)

        for example in self.examples:
            grade, messages = self.grade_example(code_obj, example)
            if messages is None:
                out = min(grade, grade)
            else:
                return grade, messages
        return out, messages

    def interact(self, code_obj, inputs):
        namespace = self.namespace()
        inputs = list(reversed(inputs))
        interaction = []

        @functools.wraps(print)
        def print_func(*args, **kwds):
            data = print_str(*args, **kwds)
            if interaction and isinstance(interaction[-1], Out):
                data = interaction.pop() + data
            interaction.append(Out(data))

        @functools.wraps(input)
        def input_func(msg=None):
            if msg:
                interaction.append(Prompt(msg))
            try:
                data = inputs.pop()
            except IndexError:
                raise errors.DidNotTerminateError
            interaction.append(In(data))
            return data

        backup_print_input = print, input
        __builtins__['print'] = print_func
        __builtins__['input'] = input_func

        # Run code
        try:
            exec(code_obj, namespace)
        finally:
            __builtins__['print'], __builtins__['input'] = backup_print_input
        return interaction

    def grade_example(self, code_obj, example):
        inputs = [str(x) for x in example if x.is_input]

        try:
            interaction = self.interact(code_obj, inputs)
        except Exception as ex:
            return Decimal(0), message_from_exception(ex)
        else:
            if self.is_equivalent(interaction, example):
                return Decimal(100), None
            else:
                return Decimal(0), message_bad_format(interaction, example)

    def is_equivalent(self, interaction, example):
        '''Return True if the two interactions are equivalent.

        Normally it is used to compare the result of running a program with
        the expected example template.
        '''

        if len(interaction) != len(example):
            return False
        else:
            if interaction[:-1] != example[:-1]:
                return False
            elif interaction[-1].strip() != example[-1].strip():
                return False
            return True


if __name__ == '__main__':
    examples = parse_io_template('-->"hello world"')
    answer = 'print(print)'

    g = PyIOGraderTemplate(examples, answer=answer)
    print(g.grade("msg = 'hello world'\nprint(msg)"))
    print(g.grade("print('hi world')"))

    #
    #
    #
    #
    #
    #
    #
    #
    #
    #
    #
    #
    #
    #
    answer = '''
def soma(x, y):
    return x + y
'''
    tests = '''
assert 'soma' in globals(), 'Não apague ou renomeie a função soma()'
assert_equal(soma(1, 2), 3)
assert_similar(soma(1.0, 2.0), 3.0)
for x in [0, 1.0, 2, -1, -10]:
    for y in [0, 1.0, 2, -1, -10]:
        assert_similar(ANSWER.soma(x, y), soma(x, y))
'''



from .util import indent


class GradingError(Exception):

    '''Base class for all errors that may occur during grading.'''

    def message(self, format='text'):  # @ReservedAssignment
        '''Return a formatted message string for the given error'''

        try:
            method = getattr(self, 'message_' + format)
        except AttributeError:
            raise ValueError('invalid format: %r' % format)
        return method()

    def message_text(self):
        '''Message formatted as plain text'''

        return repr(self)

    def name(self):
        '''Return the error name'''
        return type(self).__name__

    def title(self):
        '''Return the error title'''

        try:
            return self.TITLE
        except AttributeError:
            name = self.name()
            if name.endswith('Error'):
                name = name[:-5]
            chars = [(c if c.islower() else ' ' + c) for c in name]
            return ''.join(chars)


class BuildError(GradingError):

    '''Flags a syntax error in the input code or any kind of error that may
    prevent the code from building correctly.'''


class UnexpectedError(GradingError):

    '''Code terminates with a segfault, uncaught exception, or any other
    non-zero exit code.'''


class TimeExceededError(GradingError):

    '''Code does not terminate until the maximum estipulated runtime is
    reached.'''


class UnusedInputError(GradingError):

    '''Code runs without errors but do not uses all input values expected to
    be used. Do not check if the output values written up to that point are
    correct.'''


class MissingInputError(GradingError):

    '''Program expects a non-existing input value.'''


#
# Wrong answers
#
class WrongAnswerError(GradingError, ValueError):

    '''Code runs without error, but computes wrong output values.'''

    def __init__(self, wrong, right):
        self.wrong = wrong
        self.right = right
        self.args = (wrong, right)

    def message_text(self):
        title = self.title()
        msg = [
            '-' * len(title),
            title,
            '-' * len(title),
            '\nExpected answer:',
            indent(self.right.format('text')),
            '\nObtained:',
            indent(self.wrong.format('text')),
        ]
        return '\n'.join(msg) + '\n'


class PresentationError(WrongAnswerError):

    '''Code runs without error and probably computes the correct values.
    However it does not format the output correctly.'''


class BadPromptError(WrongAnswerError):
    '''Code that would be 100% correct if the texts in the prompts were
    ignored.'''
